# Copyright 2021 Benedict Suska <benedict.suska@teckids.org>
# Copyright 2021 Dominik George <dominik.george@teckids.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import time
from collections import OrderedDict
from machine import Pin, PWM
from credentials import *
from wifi import Wifi
from mytrix import Matrix
from random import getrandbits

try:
    from mmllib.esp import play
    from mmllib.mml import mml
except ImportError:
    mml = None


def set_colour(matches, event):
    """Set the RGB LED colour depending on the colour name in a message."""
    colour = matches.group(1).lower().strip()
    sender = event["sender"]

    # Mapping colour names to RGB values
    colour_map = {
        "schwarz": (0, 0, 0),
        "blau": (0, 0, 255),
        "rot": (255, 0, 0),
        "gruen": (0, 255, 0),
        "tuerkis": (0, 255, 255),
        "lila": (139, 0, 139),
        "orange": (255, 140, 0),
        "pink": (255, 105, 180),
        "rosa": (255, 0, 255),
        "weiss": (255, 255, 255),
    }

    if is_allowed(sender):
        if not colour in colour_map:
            # The colour name is not in the mapping
            matrix.send_room_message(
                ROOM_ID, "Diese Farbe kenne ich leider nicht, %s :(." % (sender,)
            )
            return

        # Map to red, blue and green values
        r, g, b = colour_map[colour]
        set_leds(r, g, b)

        matrix.send_room_message(ROOM_ID, "Ich leuchte jetzt %s, %s!" % (colour, sender))


def is_allowed(sender):
    """Check whether a sender is allowed to control the cheese."""
    with open("users.txt", "rt") as f:
        users = f.readlines()

    if "%s\n" % (sender,) in users:
        return True
    else:
        matrix.send_room_message(ROOM_ID, "Du hast mir gar nichts zu sagen, %s!" % (sender,))
        return False


def allow_user(matches, event):
    """Add a user to the allow list."""
    user = matches.group(1)
    sender = event["sender"]

    if is_allowed(sender):
        with open("users.txt", "at") as f:
            f.write(user + "\n")
        matrix.send_room_message(ROOM_ID, "%s darf mich jetzt leuchten lassen!" % (user,))


def not_stupid(matches, event):
    """Say that the cheese is not stupid."""
    sender = event["sender"]
    matrix.send_room_message(ROOM_ID, "Ich bin nicht bloed, %s!" % (sender,))
    blink((255, 0, 0), 5, 0.3)


def party(matches, event):
    """Blink wildly."""
    sender = event["sender"]
    if is_allowed(sender):
        matrix.send_room_message(
            ROOM_ID, "Cheesy party, dingelingeling! Chessy-cheesy-cheeeeese, yay!"
        )
        for _ in range(10):
            r = getrandbits(8)
            g = getrandbits(8)
            b = getrandbits(8)
            blink((r, g, b), 2, 0.3)


def off(matches, event):
    """Turn LEDs off."""
    sender = event["sender"]
    if is_allowed(sender):
        matrix.send_room_message(ROOM_ID, "Nerv ich? Na gut...")
        set_leds(0, 0, 0)


def music(matches, event):
    """Play MML music."""
    sender = event["sender"]
    if is_allowed(sender):
        if mml is not None:
            matrix.send_room_message(ROOM_ID, "La la la...")
            play(mml(matches.group(1)), pin_buzzer)
        else:
            matrix.send_room_message(ROOM_ID, "Ich kann nicht singen :(, %s..." % (sender,))


def fallback(matches, event):
    """Fallback fo unknown messages."""
    sender = event["sender"]
    matrix.send_room_message(ROOM_ID, "Was willst du, %s?" % (sender,))


def set_leds(r, g, b):
    """Set PWM duty (scaled from 256 to 1024)."""
    led_red.duty(r * 4)
    led_green.duty(g * 4)
    led_blue.duty(b * 4)


def blink(rgb, times, pause):
    """Blink in rgb color, times times, with pause ms delay."""
    r, g, b = rgb
    for _ in range(times):
        set_leds(r, g, b)
        time.sleep(pause)
        set_leds(0, 0, 0)
        time.sleep(pause)


# Define LED colour pins and PWM objects
pin_red = Pin(13, Pin.OUT)
pin_green = Pin(12, Pin.OUT)
pin_blue = Pin(14, Pin.OUT)
led_red = PWM(pin_red, 5000)
led_green = PWM(pin_green, 5000)
led_blue = PWM(pin_blue, 5000)

# Define buzzer pin number
pin_buzzer = 2

# Connect to wifi
network = Wifi(SSID, PSK)
network.connect()

# Connect to Matrix server
matrix = Matrix(HOMESERVER, MATRIX_ID, ACCESS_TOKEN)

# On which texts to react and which LEDs to turn on
cases = OrderedDict()
cases["^%s, leuchte (.*)" % (NICK,)] = set_colour
cases["^%s, hoer auf (.*)" % (NICK,)] = allow_user
cases["^%s, aus" % (NICK,)] = off
cases["^%s, spiele (.*)" % (NICK,)] = music
cases["bloeder kaese"] = not_stupid
cases["kaeseparty"] = party
cases["^%s.*" % (NICK,)] = fallback

if network.is_connected():
    # Set my nickname and avatar
    matrix.set_displayname(NICK)
    matrix.set_avatar(AVATAR)

    # Join the room
    matrix.join_room(ROOM_ID)

    # Send message that I am there
    matrix.send_room_message(ROOM_ID, "Hey! Alles cheesy, ich bin bereit zum Leuchten!")
    matrix.send_room_message(ROOM_ID, "Lass mir etwas Zeit, auf deine Nachrichten zu reagieren!")

    # Wait for a message
    while True:
        matrix.react_room_messages(ROOM_ID, cases, regex=True)
        time.sleep(3)
